/*
 * Maxim Domentii, 333CC
 */

#include "lcd.h"

#include <avr/io.h>
#include <util/delay.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* LCD specific stuff. */
#define FST_LINE_FST_CHAR 0x00
#define SND_LINE_FST_CHAR 0x40
#define EMPTY_LINE "                "

/*Defines a macro for the keypad.h header file. 
KEYPAD_PORT is the microcontroller PORT Register to which 4X4 keypad is connected. 
Default PORT Register in keypad.h header file is PORTB*/
#define	KEYPAD_PORT	PORTB

/*Defines a macro for the keypad.h header file. 
KEYPAD_PIN is the microcontroller PIN Register to which 4X4 keypad is connected. 
Default PIN Register in keypad.h header file is PINB*/
#define	KEYPAD_PIN	PINB

#include "keypad.h"

void LCD_clear (void) {
    LCD_printAt (FST_LINE_FST_CHAR, EMPTY_LINE);
    LCD_printAt (SND_LINE_FST_CHAR, EMPTY_LINE);
}

void button_init (void) {
	/*PortB's upper 4 bits are declared input and lower 4 bits are declared output(4x4 Keypad is connected)*/
    DDRB=0x0f;

	/*PortB's lower 4 bits are given high value and pull-up are enabled for upper 4 bits*/
	PORTB=0xff;
	
	/* Button BTN from board */
	DDRD &= ~(1 << PB6);
	PORTD |= (1 << PB6);
	
	/* Led from board */
	DDRD |= (1 << PD7);
	PORTD &= ~(1 << PD7);
}

/* Convert read_keypad() result to char */
char getChar(unsigned char keypad_value){
	switch(keypad_value)
	{
		case 0:
			return '0';

		case 1:
			return '1';
			
		case 2:
			return '2';

		case 3:
			return '3';

		case 4:
			return '4';

		case 5:
			return '5';

		case 6:
			return '6';

		case 7:
			return '7';

		case 8:
			return '8';

		case 9:
			return '9';

		case 10:
			return '.';

		case 11:
			return '-';

		case 12:
			return 'A';

		case 13:
			return 'B';

		case 14:
			return 'C';

		case 15:
			return 'D';
	}
	return '\0';
}

/* Read a number until 'A' is read */
char* readNumber(int num_length){
	char *num = strdup ("");;
	int k = 0;
	unsigned char keypad_value;
	char aux[2];
	
	while (k <= num_length ){
		keypad_value=read_keypad();
		
		if (getChar(keypad_value) == 'A'){
			if (k == 0)
				return "0";
			break;
		}
		if (keypad_value!=0xff){
			
			/* Minus can be only at the begining */
			if (getChar(keypad_value) == '-' && k != 0){}
			else{
				
				/* . can apear in a number only once */
				if (getChar(keypad_value) == '.' && strchr(num, '.') !=NULL){}
				else{
					aux[0] = getChar(keypad_value);
					aux[1] = '\0';
					num = strcat(num, aux);
					k++;
				}
			}
		}
		else;
		_delay_ms(200);
	}
	
	return num;
}

/* Convert an int to string */
char* intoa(int i){
	char *s = malloc(4*sizeof(char));
	sprintf(s,"%d",i);
	return s;
}

/* Convert a double to string */
char* fltoa(double i){
	char *s = malloc(4*sizeof(char));
	sprintf(s,"%.2f",i);
	return s;
}

double** readA(int dim){
	int i,j;
	char *line = malloc(16*sizeof(char));
	char *elem = malloc(16*sizeof(char));
	char *num;
	
	double **A = malloc(dim*sizeof(double*));
	for (i=0;i<dim;i++){
		A[i] = calloc(dim,sizeof(double));
	}
	
	for (i=0;i<dim;i++){
		LCD_clear();
		sprintf(line, "Line %d", i);
		LCD_printAt (FST_LINE_FST_CHAR, line);
		for(j=0;j<dim;j++){
			LCD_printAt (SND_LINE_FST_CHAR, EMPTY_LINE);
			sprintf(elem, "Elem %d: ", j);
			LCD_printAt (SND_LINE_FST_CHAR, elem);
			
			num = readNumber(7);
			A[i][j] = atof(num);
			
			LCD_printAt (SND_LINE_FST_CHAR+12, num);
			_delay_ms(1000);
		}
	}
	LCD_clear();
		
	return A;
}

double* readb(int dim){
	int i;
	char *elem = malloc(16*sizeof(char));
	char *num;
	
	double *b = calloc(dim,sizeof(double));
		
	for(i=0;i<dim;i++){
		LCD_printAt (SND_LINE_FST_CHAR, EMPTY_LINE);
		sprintf(elem, "Elem %d: ", i);
		LCD_printAt (SND_LINE_FST_CHAR, elem);
		
		num = readNumber(7);
		b[i] = atof(num);
		
		LCD_printAt (SND_LINE_FST_CHAR+12, num);
		_delay_ms(1000);
	}
	LCD_clear();
		
	return b;
}

void writeResult(double *x, int dim){
	int i;
	char *elem = malloc(16*sizeof(char));
	char *num;
	
	for(i=0;i<dim;i++){
		LCD_printAt (SND_LINE_FST_CHAR, EMPTY_LINE);
		sprintf(elem, "x%d: ", i);
		LCD_printAt (SND_LINE_FST_CHAR, elem);
		
		num = fltoa(x[i]);
		
		LCD_printAt (SND_LINE_FST_CHAR+12, num);
		_delay_ms(3000);
	}
	LCD_clear();
}

double powd(double x, double y){
	double sum;
	int i;
	if (y==0)
		return 1;

	sum=x;
	
	for (i=1; i<=y-1; i++)
		sum=sum*x;

	return(sum);
}

/* Recursive definition of determinate using expansion by minors */
double determinant(double **a,int n){
   int i,j,j1,j2;
   double det = 0;
   double **m = NULL;

   if (n < 1) { /* Error */

   } else if (n == 1) { /* Shouldn't get used */
      det = a[0][0];
   } else if (n == 2) {
      det = a[0][0] * a[1][1] - a[1][0] * a[0][1];
   } else {
      det = 0;
      for (j1=0;j1<n;j1++) {
         m = malloc((n-1)*sizeof(double *));
         for (i=0;i<n-1;i++)
            m[i] = malloc((n-1)*sizeof(double));
         for (i=1;i<n;i++) {
            j2 = 0;
            for (j=0;j<n;j++) {
               if (j == j1)
                  continue;
               m[i-1][j2] = a[i][j];
               j2++;
            }
         }
         det += powd(-1.0,j1+2.0) * a[0][j1] * determinant(m,n-1);
         for (i=0;i<n-1;i++)
            free(m[i]);
         free(m);
      }
   }
   return(det);
}

/* Find the cofactor matrix of a square matrix */
double** co_factor(double **a,int n){
   int i,j,ii,jj,i1,j1;
   double det;
   double **c, **b;

   c = malloc(n*sizeof(double *));
   for (i=0;i<n;i++)
     c[i] = malloc(n*sizeof(double));
 
   b = malloc(n*sizeof(double *));
   for (i=0;i<n;i++)
     b[i] = malloc(n*sizeof(double));

   for (j=0;j<n;j++) {
      for (i=0;i<n;i++) {

         /* Form the adjoint a_ij */
         i1 = 0;
         for (ii=0;ii<n;ii++) {
            if (ii == i)
               continue;
            j1 = 0;
            for (jj=0;jj<n;jj++) {
               if (jj == j)
                  continue;
               c[i1][j1] = a[ii][jj];
               j1++;
            }
            i1++;
         }

         /* Calculate the determinate */
         det = determinant(c,n-1);

         /* Fill in the elements of the cofactor */
         b[i][j] = powd(-1.0,i+j+2.0) * det;
      }
   }
   for (i=0;i<n-1;i++)
      free(c[i]);
   free(c);
   
   return b;
}

/* Transpose of a square matrix, do it in place */
void transpose(double **a,int n){
   int i,j;
   double tmp;

   for (i=1;i<n;i++) {
      for (j=0;j<i;j++) {
         tmp = a[i][j];
         a[i][j] = a[j][i];
         a[j][i] = tmp;
      }
   }
}

double** inv(double **A, int dim){
	int i,j;
	
	double det = determinant(A, dim);
	
	/* Check det to be != 0 */
	if (det == 0 || det == 0.0 || det == -0 || det == -0.0 ){
		return NULL;
	}
	
	A = co_factor(A, dim);
	transpose(A, dim);
	
	for (i=0; i<dim; i++){
		for (j=0; j<dim; j++){
			A[i][j] = A[i][j] / det;
		}
	}
	
	return A;
}

double vectors_dot_prod(double *x, double *y, int n){
    double res = 0.0;
    int i;
    for (i = 0; i < n; i++){
        res += x[i] * y[i];
    }
    return res;
}

double* matrix_vector_mult(double **mat, double *vec, int n){
	double *result = malloc(n * sizeof(double));
    int i; 
	
	for (i = 0; i < n; i++){
        result[i] = vectors_dot_prod(mat[i], vec, n);
    }
	return result;
}

int main(){
	/* Var declaration */
	char *s_dim;
	int i_dim;
	double **A;
	double *b;
	double *result;
	
    //LCD
    LCD_init();
	button_init();
	
	while(1){
		LCD_clear();
		
		/* Read dimension*/
		LCD_printAt (FST_LINE_FST_CHAR, "System dim: ");
		
		s_dim = readNumber(4);
		i_dim = atoi(s_dim);
		
		/* Check dimension. Must be positive integer */
		if (i_dim <= 0){
			LCD_printAt (FST_LINE_FST_CHAR+12, intoa(i_dim));
			LCD_printAt (SND_LINE_FST_CHAR, "Dim must be pos!");
			_delay_ms(3000);
			continue;
		}
		
		LCD_printAt (FST_LINE_FST_CHAR+12, intoa(i_dim));
		_delay_ms(1000);
		LCD_clear();
		
		/* Read matrix A */
		LCD_printAt (FST_LINE_FST_CHAR, "Read matrix A...");
		_delay_ms(1000);
		LCD_clear();
		
		A = readA(i_dim);
		
		/* Read vector b */
		LCD_printAt (FST_LINE_FST_CHAR, "Read vector b...");
		_delay_ms(1000);
		
		b = readb(i_dim);
		
		/* Resolve system */
		A  = inv(A, i_dim);
		
		if (A == NULL){
			LCD_printAt (FST_LINE_FST_CHAR, "Det. of inv. = 0");
			_delay_ms(3000);
		}
		else{		
			result = matrix_vector_mult(A, b, i_dim);
		
			/* Write results x */
			LCD_printAt (FST_LINE_FST_CHAR, "Results...");
			_delay_ms(3000);
		
			writeResult(result, i_dim);
		}		
	}
	
	return 0;
}

